def seed_everything(seed=1234):
    '''Set the seed for :mod:`torch`, :mod:`random`, :mod:`numpy` and :mod:`torch.cuda`, as well as
    the ``PYTHONHASHSEED`` env var.

    .. warning::

        It seems that some amount of indeterminism persists when using CUDA. While results are
        reproducible on the CPU, small deviations are introduced by the GPU. My theory based on
        second-hand information is that the order of reductions operations in the context of CUDA
        threads is not really deterministic. Since floating point math is not perfectly associative,
        the order with which reductions are performed on the individual threads's outputs is not
        guaranteed.

    Parameters
    ----------
    seed    :   int
                Seed value to use.
    '''
    import random, torch, os, numpy as np       # noqa
    # Manually seeding everything does not work completely on the GPU
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)


def make_fill_polygons(xs, ys, zs):
    '''Make a set of polygons to fill the space below a line in 3d.

    Returns
    -------
    mpl_toolkits.mplot3d.art3d.Poly3DCollection
    '''
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    v = []
    h = 0
    for k in range(0, len(xs) - 1):
        x = [xs[k], xs[k+1], xs[k+1], xs[k]]
        y = [ys[k], ys[k+1], ys[k+1], ys[k]]
        z = [zs[k], zs[k+1],       h,     h]
        v.append(list(zip(x, y, z)))
    poly3dCollection = Poly3DCollection(v)
    return poly3dCollection


def available_optimizers():
    '''List names of all available torch optimizers form :py:mod:`torch.optim`.

    Returns
    -------
    list(str)
        A list of all optimizer names.
    '''
    import torch
    # get all module properties which aren't magic
    available_optimizers = set(getattr(torch.optim, name) for name in dir(torch.optim) if
                               not name.startswith('__'))
    # remove everything which is a module and not a class (looking at you, lr_scheduler o_o)
    available_optimizers = filter(lambda o: isinstance(o, type), available_optimizers)
    # map them to their class name (w/o module)
    available_optimizers = map(lambda o: o.__name__, available_optimizers)
    return list(available_optimizers)


def create_optimizer(model, name, **kwargs):
    '''Create an optimizer for ``model``\ s parameters. Will disregard all params
    with ``requires_grad == False``.

    Parameters
    ----------
    model   :   torch.nn.Module
    name    :   str
                Name of the optimizer
    kwargs  :   dict
                All arguments which should be passed to the optimizer.

    Raises
    ------
    ValueError
        If superfluous ``kwargs`` are passed.

    '''
    import torch
    if name not in available_optimizers():
        raise ValueError(f'Unknown optimizer {name}')

    params = [p for p in model.parameters() if p.requires_grad]
    return getattr(torch.optim, name)(params, **kwargs)


def initialize_model(module, bias_val=0.01):
    '''Perform weight initialization on `module`. This is somewhat hacky since
    it assumes the presence of `weight` and/or `bias` fields on the module. Will
    skip if not present.

    Parameters
    ----------
    module  :   torch.nn.Module
                The model
    bias_val    :   float
                    Constant for biases (should be small and positive)

    Raises
    ------
    ValueError
        If ``module`` is not one of the known models (currently :class:`~ikkuna.models.AlexNetMini`
        and :class:`~ikkuna.models.DenseNet`)
    '''
    from ikkuna import models
    import torch.nn as nn
    if isinstance(module, models.AlexNetMini):
        for m in module.modules():
            if hasattr(module, 'weight'):
                nn.init.xavier_uniform_(module.weight)
            if hasattr(module, 'bias'):
                module.bias.data.fill_(bias_val)
    elif isinstance(module, models.DenseNet):
        # Official init from torch repo.
        for m in module.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight.data)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                m.bias.data.zero_()
    else:
        raise ValueError(f'Don\'t know how to initialize {module.__class__.__name__}')


import subprocess


def get_memory_stats(mode='total', unit='mb'):
    '''Get the total|used|free gpu memory.

    Parameters
    ----------
    mode    :   str
                One of ``'total'``, ``'used'`` or ``'free'``

    Returns
    -------
    dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    '''
    if mode not in ['total', 'used', 'free']:
        raise ValueError

    units = ['b', 'kb', 'mb', 'gb']
    factors = [1024 ** i for i in range(len(units))]
    if unit not in units:
        raise ValueError(f'Unknown unit "{unit}"')

    memory = subprocess.check_output(
        [
            'nvidia-smi', f'--query-gpu=memory.{mode}',
            '--format=csv,nounits,noheader'
        ], encoding='utf-8')
    # Convert lines into a dictionary
    index = units.index(unit)
    # nvidia-smi gives mb, so go back to b and then up to the selected unit
    size = [int(x) * factors[2] / factors[index] for x in memory.strip().split('\n')]
    return dict(zip(range(len(size)), size))


import torchvision
from torchvision.transforms import ToTensor, Compose
import numpy as np
import torch


def load_dataset(name, train_transforms=None, test_transforms=None):
    '''Retrieve a dataset and determine the number of classes. This estimate is
    obtained from the number of different values in the training labels.

    Parameters
    ----------
    name    :   str
                Dataset name in :mod:`torchvision.datasets`

    Returns
    -------
    tuple
        2 :class:`~train.DatasetMeta` s are returned, one for train and one test set
    '''
    from train import DatasetMeta
    train_transforms = Compose(train_transforms) if train_transforms else ToTensor()
    test_transforms  = Compose(test_transforms) if test_transforms else ToTensor()
    try:
        dataset_cls   = getattr(torchvision.datasets, name)
        dataset_train = dataset_cls('/home/share/data',
                                    download=True,
                                    train=True,
                                    transform=train_transforms
                                    )
        dataset_test  = dataset_cls('/home/share/data',
                                    download=True,
                                    train=False,
                                    transform=test_transforms
                                    )
    except AttributeError:
        raise NameError(f'Dataset {name} unknown.')

    def num_classes(dataset):
        if dataset.train:
            _, labels = dataset.train_data, dataset.train_labels  # noqa
        else:
            _, labels = dataset.test_data, dataset.test_labels  # noqa

        # infer number of classes from labels. will fail if not all classes occur in labels
        if isinstance(labels, np.ndarray):
            return np.unique(labels).size()
        elif isinstance(labels, list):
            return len(set(labels))
        elif isinstance(labels, torch.Tensor):
            return labels.unique().numel()
        else:
            raise ValueError(f'Unexpected label storage {labels.__class__.__name__}')

    def shape(dataset):
        if dataset.train:
            data, _ = dataset.train_data, dataset.train_labels  # noqa
        else:
            data, _ = dataset.test_data, dataset.test_labels  # noqa

        # if only three dimensions, assume [N, H, W], else [N, H, W, C]
        N, H, W = data.shape[:3]
        C = data.shape[-1] if len(data.shape) == 4 else 1
        return (N, H, W, C)

    meta_train = DatasetMeta(dataset=dataset_train, num_classes=num_classes(dataset_train),
                             shape=shape(dataset_train))
    meta_test  = DatasetMeta(dataset=dataset_test, num_classes=num_classes(dataset_test),
                             shape=shape(dataset_test))

    return meta_train, meta_test



'''
.. moduleauthor:: Rasmus Diederichsen

This module contains the base definition for subscriber functionality. The
:class:`~ikkuna.export.subscriber.Subscriber` class should be subclassed for adding new metrics.

'''
import abc
from collections import defaultdict
import ikkuna.visualization
from ikkuna.export.messages import MessageBundle, TrainingMessage


class Subscription(object):
    '''Specification for a subscription that can span multiple kinds and a tag.

    Attributes
    ----------
    _tag    :   str
                Tag for filtering the processed messages
    _subscriber :   ikkuna.export.subscriber.Subscriber
                    The subscriber associated with the subscription
    counter :   dict(ikkuna.utils.NamedModule or str, int)
                Number of times the subscriber was called for each module label or meta data
                identifier. Since one :class:`ikkuna.export.subscriber.Subscriber` is associated
                with only one configuration of :class:`ikkuna.export.messages.MessageBundle`, this
                will enable proper subsampling of message streams.
    kinds   :   list(str)
                List of string identifiers for different message kinds. These are all the
                message kinds the subscriber wishes to receive
    _subsample  :   int
                    Factor for subsampling incoming messages. Only every ``subsample``-th
                    message will be processed.
    '''

    def __init__(self, subscriber, kinds, tag=None, subsample=1):
        '''
        Parameters
        ----------
        subscriber  :   ikkuna.export.subscriber.Subscriber
                        Object that wants to receive the messages
        tag :   str or None
                Optional tag for filtering messages. If ``None``, all messages will be
                relayed
        '''
        self._tag        = tag
        self._subscriber = subscriber
        self._kinds      = kinds
        self._counter    = defaultdict(int)
        self._subsample  = subsample

    @property
    def counter(self):
        # caution: if you alter this dict, you're on your own
        return self._counter

    @property
    def kinds(self):
        return self._kinds

    def _handle_message(self, message):
        '''Process a newly arrived message. Subclasses should override this method for any special
        treatment.

        Parameters
        ----------
        message    :   ikkuna.export.messages.Message
        '''
        data = MessageBundle(message.key, message.kind)
        data.add_message(message)
        self._subscriber.process_message_bundle(data)

    def handle_message(self, message):
        '''Callback for receiving an incoming message.

        Parameters
        ----------
        message    :   ikkuna.export.messages.TrainingMessage
        '''
        if not (self._tag is None or self._tag == message.tag):
            return

        if message.kind not in self.kinds:
            return

        if isinstance(message, TrainingMessage):
            key = (message.module, message.kind)
        else:
            key = message.kind
        if self._counter[key] % self._subsample == 0:
            self._handle_message(message)
        self._counter[key] += 1


class SynchronizedSubscription(Subscription):
    '''A subscription which buffers messages and publishes a set of messages, each of a different
    kind, when one round (a train step) is over. This is useful for receiving several kinds of
    messages in each train step and always have them be processed together.'''

    def __init__(self, subscriber, kinds, tag=None, subsample=1):
        super().__init__(subscriber, kinds, tag, subsample)
        self._current_seq = None
        self._identifiers = {}
        self._step        = None

    def _new_round(self, seq):
        '''Start a new round of buffering, clearing the previous cache and resetting the record for
        which kinds were received in this round.

        Parameters
        ----------
        seq :   int
                Sequence number for the new round

        Raises
        ------
        RuntimeError
            If not all desired kinds have been received for all identifiers yet in the current round
        '''
        for bundle in self._identifiers.values():
            if not bundle.complete():
                raise RuntimeError(f'Bundle for module {bundle._module} not yet complete.')
        self._current_seq = seq
        self._identifiers = {}

    def _publish_complete(self):
        delete_these = []
        # any full? publish
        for identifier, message_bundle in self._identifiers.items():
            if message_bundle.complete():
                self._subscriber.process_message_bundle(message_bundle)
                delete_these.append(identifier)

        # purge published data
        for identifier in delete_these:
            del self._identifiers[identifier]

    def _handle_message(self, message):
        '''Start a new round if a new sequence number is seen.'''

        # if we get a new sequence number, a new train step must have begun
        if self._current_seq is None or self._current_seq != message.seq:
            self._new_round(message.seq)

        # module not seen -> init data
        key = message.key
        if key not in self._identifiers:
            self._identifiers[key] = MessageBundle(key, self.kinds)
        self._identifiers[key].add_message(message)

        self._publish_complete()


class Subscriber(abc.ABC):
    '''Base class for receiving and processing activations, gradients and other stuff into
    insightful metrics.'''

    def __init__(self, subscription):
        '''
        Parameters
        ----------
        subscription    :   Subscription
        '''
        self._subscription = subscription

    @abc.abstractmethod
    def _metric(self, message_or_data):
        '''This is where the magic happens. Subclasses should override this method so that they can
        compute their metric upon reception of their desired messages. They should then use their
        :attr:`~ikkuna.export.subscriber.PlotSubscriber.backend` property to publish the metric (if
        they display line plots).

        Parameters
        ----------
        message_or_data :   ikkuna.export.messages.Message
                            Can either be :class:`~ikkuna.export.messages.MetaMessage` if the
                            Subscriber is not interested in actual training artifacts, or
                            :class:`~ikkuna.export.messages.TrainingMessage`
        '''
        pass

    def receive_message(self, message):
        '''Process a message received from an :class:`~ikkuna.export.Exporter`.'''
        self._subscription.handle_message(message)

    def process_message_bundle(self, message_bundle):
        '''Callback for processing a :class:`~ikkuna.export.messages.MessageBundle` object with
        :class:`~ikkuna.export.messages.TrainingMessage`\ s or
        :class:`~ikkuna.export.messages.TrainingMessage`\ s in it.

        Parameters
        ----------
        message_bundle    :   ikkuna.export.messages.MessageBundle
                            The exact nature of this package is determined by the
                            :class:`ikkuna.export.subscriber.Subscription` attached to this
                            :class:`ikkuna.export.subscriber.Subscriber`.

        Raises
        ------
        ValueError
            If the received :class:`~ikkuna.export.messages.MessageBundle` object is not
            :meth:`~ikkuna.export.messages.MessageBundle.complete()`
        '''
        if not message_bundle.complete():
            raise ValueError(f'Data received for "{message_bundle._module}" is not complete.')

        self._metric(message_bundle)


class PlotSubscriber(Subscriber):
    '''Base class for subscribers that output scalar or histogram values per time and module

    Attributes
    ----------
    _backend    :   ikkuna.visualization.Backend
                    Plotting backend
    '''

    def __init__(self, subscription, plot_config, backend='tb', **tbx_params):
        '''
        Parameters
        ----------
        ylims   :   tuple(int, int)
                    Optional Y-axis limits
        plot_config :   dict
                        Configuration parameters for plotting. Relevant keys are ``title``,
                        ``xlabel``, ``ylabel`` and ``ylims``. Which of them are actually used
                        depends on the :class:`~ikkuna.visualization.Backend`
        **tbx_params    :   dict
                            Keywords for the :class:`tensorboardX.SummaryWriter`
        '''
        super().__init__(subscription)

        self._backend = ikkuna.visualization.get_backend(backend, plot_config, **tbx_params)

    @property
    def backend(self):
        '''ikkuna.visualization.Backend: The backend to use for plotting'''
        return self._backend

    @abc.abstractmethod
    def _metric(self, message_or_data):
        pass



from .alexnetmini import AlexNetMini
from .densenet import DenseNet
from .resnet import ResNet
from .resnet import resnet18, resnet34, resnet50, resnet101, resnet152

__all__ = ['AlexNetMini', 'DenseNet', 'ResNet']



import torch
import numpy as np

from ikkuna.export.subscriber import PlotSubscriber, SynchronizedSubscription

ZERO_TENSOR = torch.tensor(0.0).cuda()


class RatioSubscriber(PlotSubscriber):
    '''A :class:`~ikkuna.export.subscriber.Subscriber` which computes the average ratio between two
    quantities.  The dividend will be the first element of the
    :attr:`~ikkuna.export.subscriber.Subscription.kinds` property, the divisor the second.
    Therefore it is vital to pass the message kinds to the
    :class:`~ikkuna.export.subscriber.Subscription` object in the correct order.'''

    def __init__(self, kinds, tag=None, subsample=1, ylims=None, backend='tb', absolute=True,
                 **tbx_params):
        '''
        Parameters
        ----------
        absolute :  bool
                    Whether to use absolute ratio
        '''
        title        = f'{kinds[0]}/{kinds[1]} ratios per layer'
        ylabel       = 'Ratio'
        xlabel       = 'Train step'
        subscription = SynchronizedSubscription(self, kinds, tag, subsample)
        super().__init__(subscription, {'title': title,
                                        'ylabel': ylabel,
                                        'ylims': ylims,
                                        'xlabel': xlabel},
                         backend=backend, **tbx_params)
        if absolute:
            self._metric_postprocess = torch.abs
        else:
            self._metric_postprocess = lambda x: x

    def _metric(self, message_bundle):
        '''The ratio between the two kinds is computed over the subset of not-NaN values and added
        to the record.'''

        module_name  = message_bundle.identifier

        dividend = message_bundle.data[self._subscription.kinds[0]]
        divisor  = message_bundle.data[self._subscription.kinds[1]]

        ######################################################################################
        #  We need to see how many NaNs we have and compute the mean only over the non-nans  #
        ######################################################################################
        ratio_tensor = self._metric_postprocess(dividend.div(divisor))
        n            = float(divisor.numel())
        nan_tensor   = torch.isnan(ratio_tensor)
        n_nans       = nan_tensor.sum().to(torch.float32)
        if n_nans > 0:
            ratio_sum = torch.where(1 - nan_tensor, ratio_tensor, ZERO_TENSOR).sum()
        else:
            ratio_sum = ratio_tensor.sum()
        ratio = (ratio_sum / (n - n_nans)).item()

        if np.isnan(ratio):
            raise ValueError(f'NaN value ratio for {module_name}')

        self._backend.add_data(module_name, ratio, message_bundle.seq)



'''
.. moduleauthor Rasmus Diederichsen

This module defines :class:`ikkuna.models.AlexNetMini`, a reduced version of AlexNet. Adapted from
:meth:`torchvision.models.alexnet()`.
'''
import torch


class AlexNetMini(torch.nn.Module):     # Pytorch Sphinx-doc is buggy here, so use full path
    '''Reduced AlexNet (basically just a few conv layers with relu and
    max-pooling) which attempts to adapt to arbitrary input sizes, provided they are large enough to
    survive the strides and conv cutoffs.

    Attributes
    ---------
    features    :   torch.nn.Module
                    Convolutional module, extracting features from the input
    classifier  :   torch.nn.Module
                    Classifier with relu and dropout
    H_out   :   int
                Output height of the classifier
    W_out   :   int
                Output width of the classifier
    '''
    def __init__(self, input_shape, num_classes=1000, exporter=None):
        super(AlexNetMini, self).__init__()
        exporter.set_model(self)

        # if channel dim not present, add 1
        if len(input_shape) == 2:
            input_shape.append(1)
        H, W, C = input_shape

        self.features = torch.nn.Sequential(
            torch.nn.Conv2d(C, 64, kernel_size=5, stride=2, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(kernel_size=3, stride=2),
            torch.nn.Conv2d(64, 192, kernel_size=3, padding=2),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(kernel_size=3, stride=2),
            torch.nn.Conv2d(192, 192, kernel_size=3, padding=1),
            torch.nn.ReLU(inplace=True),
        )
        self.H_out =  H // (2 * 2 * 2)
        self.W_out =  W // (2 * 2 * 2)
        self.classifier = torch.nn.Sequential(
            torch.nn.Dropout(),
            torch.nn.Linear(192 * self.H_out * self.W_out, 2048),
            torch.nn.ReLU(inplace=True),
            torch.nn.Dropout(),
            torch.nn.Linear(2048, 2048),
            torch.nn.ReLU(inplace=True),
            torch.nn.Linear(2048, num_classes),
        )

        exporter.add_modules(self)

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), 192 * self.H_out * self.W_out)
        x = self.classifier(x)
        return x



from .subscriber import Subscriber, Subscription, SynchronizedSubscription, PlotSubscriber
from .histogram import HistogramSubscriber
from .ratio import RatioSubscriber
from .spectral_norm import SpectralNormSubscriber
from .test_accuracy import TestAccuracySubscriber
from .train_accuracy import TrainAccuracySubscriber

__all__ = ['Subscriber', 'Subscription', 'SynchronizedSubscription', 'PlotSubscriber',
           'HistogramSubscriber', 'RatioSubscriber', 'SpectralNormSubscriber', 'TestAccuracySubscriber',
           'TrainAccuracySubscriber']



import torch
from torch.nn.functional import normalize

from ikkuna.export.subscriber import PlotSubscriber, Subscription


class SpectralNormSubscriber(PlotSubscriber):

    def __init__(self, kind, tag=None, subsample=1, ylims=None, backend='tb', **tbx_params):
        '''
        Parameters
        ----------
        kind    :   str
                    Message kind to compute spectral norm on. Doesn't make sense with kinds of
                    non-matrix type.


        For other parameters, see :class:`~ikkuna.export.subscriber.PlotSubscriber`
        '''
        if not isinstance(kind, str):
            raise ValueError('SpectralNormSubscriber only accepts 1 kind')
        subscription = Subscription(self, [kind], tag, subsample)

        title = f'Spectral norms of {kind} per layer'
        xlabel = 'Step'
        ylabel = 'Spectral norm'
        super().__init__(subscription,
                         {'title': title, 'xlabel': xlabel, 'ylims': ylims, 'ylabel': ylabel},
                         backend=backend, **tbx_params)
        self.u = dict()

    def _metric(self, message_bundle):
        '''The spectral norm computation is taken from the `Pytorch implementation of spectral norm
        <https://pytorch.org/docs/master/_modules/torch/nn/utils/spectral_norm.html>`_. It's
        possible to use SVD instead, but we are not interested in the full matrix decomposition,
        merely in the singular values.'''

        module_name = message_bundle.identifier
        # get and reshape the weight tensor to 2d
        weights     = message_bundle.data[self._subscription.kinds[0]]
        height      = weights.size(0)
        weights2d   = weights.reshape(height, -1)

        # buffer for power iteration (don't know what the mahematical purpose is)
        if module_name not in self.u:
            self.u[module_name] = normalize(weights2d.new_empty(height).normal_(0, 1), dim=0)

        # estimate singular values
        with torch.no_grad():
            for _ in range(3):      # TODO: Make niter parameter
                v = normalize(torch.matmul(weights2d.t(), self.u[module_name]), dim=0)
                self.u[module_name] = normalize(torch.matmul(weights2d, v), dim=0)

        norm = torch.dot(self.u[module_name], torch.matmul(weights2d, v)).item()

        self._backend.add_data(module_name, norm, message_bundle.seq)



from ikkuna.export.subscriber import PlotSubscriber, SynchronizedSubscription


class TrainAccuracySubscriber(PlotSubscriber):
    ''':class:`~ikkuna.export.subscriber.Subscriber` which computes the batch accuracy.'''

    def __init__(self, tag=None, subsample=1, ylims=None, backend='tb', **tbx_params):
        '''
        For parameters see :class:`~ikkuna.export.subscriber.PlotSubscriber`
        '''
        subscription = SynchronizedSubscription(self, ['network_output', 'input_labels'], tag,
                                                subsample)

        title  = f'Train batch accuracy'
        xlabel = 'Step'
        ylabel = 'Accuracy'
        super().__init__(subscription,
                         {'title': title, 'xlabel': xlabel, 'ylims': ylims, 'ylabel': ylabel},
                         backend=backend, **tbx_params)

    def _metric(self, message_bundle):
        Y           = message_bundle.data['network_output']
        labels      = message_bundle.data['input_labels']
        predictions = Y.argmax(1)
        n_correct   = (predictions == labels).sum().item()
        accuracy    = n_correct / float(labels.numel())
        self._backend.add_data('Train batch accuracy', accuracy, message_bundle.seq)



import torch


class CompoundModule(torch.nn.Module):
    '''A module comprised of more submodules which can be treated as a whole.'''

    def __init__(self, module):
        super.__init__()
        self._modules['root'] = module
        leaves = self.leaves()
        self._leaves_w_weight = set()
        self._leaves_w_bias = set()
        for m in leaves:
            if hasattr(m, 'bias') and m.bias is not None:
                self._leaves_w_bias.add(m)
            if hasattr(m, 'weight') and m.weight is not None:
                self._leaves_w_weight.add(m)

    def forward(self, x):
        return self.root(x)

    def add_module(self, *args):
        raise ValueError('Submodules don\'t make sense here.')

    def leaves(self, module=None):
        stack = list(self.children())
        children = set()
        while len(stack) > 0:
            child = stack.pop()
            for grandchild in child.children():
                if len(grandchild.children()) == 0:     # leaf
                    children.add(grandchild)
                else:
                    stack.append(grandchild)
        return children



from ikkuna.export.subscriber import PlotSubscriber, SynchronizedSubscription
from collections import defaultdict


class HistogramSubscriber(PlotSubscriber):

    '''A :class:`~ikkuna.export.subscriber.Subscriber` which subsamples training artifacts and
    computes histograms per epoch.  Histograms are non-normalized.

    Attributes
    ----------
    _buffer :   dict(list)
                Per-module buffer of values to compute more robust histogram
    '''

    def __init__(self, kinds, tag=None, subsample=1, backend='tb'):
        subscription = SynchronizedSubscription(self, kinds, tag, subsample)
        title        = f'{kinds[0]} histogram'
        ylabel       = 'Frequency'
        super().__init__(subscription, {'title': title, 'ylabel': ylabel}, backend=backend)
        self._buffer = defaultdict(list)

    def _metric(self, message_bundle):

        module_name = message_bundle.identifier
        data   = message_bundle.data[self._subscription.kinds[0]]
        self._backend.add_histogram(module_name, data, message_bundle.seq)



'''
.. data:: META_KINDS

    Message kinds which are not tied to any specific module.

.. data:: DATA_KINDS

    Message kinds which are tied to a specific module and always carry data

.. data:: ALLOWED_KINDS

    Simply the union of ``META_KINDS`` and ``DATA_KINDS``
'''

import abc
META_KINDS = {
    'batch_started', 'batch_finished', 'epoch_started', 'epoch_finished', 'input_data',
    'input_labels', 'network_output'
}

DATA_KINDS = {
    'weights', 'weight_gradients', 'weight_updates', 'biases', 'bias_gradients', 'bias_updates',
    'activations',
}

ALLOWED_KINDS = set.union(META_KINDS, DATA_KINDS)


class Message(abc.ABC):
    '''Base class for messages emitted from the :class:`~ikkuna.export.Exporter`.

    These messages are assembled into :class:`MessageBundle` objects in the
    :class:`~ikkuna.export.subscriber.Subscription`.
    '''
    def __init__(self, tag, seq, step, epoch, kind):
        self._tag  = tag
        self._seq  = seq
        self.step  = step
        self.epoch = epoch
        self.kind  = kind
        self._data = None

    @property
    def tag(self):
        ''' str: The tag associated with this message '''
        return self._tag

    @property
    def seq(self):
        '''int: Global sequence number. This counter should not reset after each epoch.'''
        return self._seq

    @property
    def step(self):
        '''int: Epoch-local sequence number (the current batch index)'''
        return self._step

    @step.setter
    def step(self, value):
        if value < 0:
            raise ValueError('Step cannot be negative.')
        else:
            self._step = value

    @property
    def epoch(self):
        '''int: Current epoch number'''
        return self._epoch

    @epoch.setter
    def epoch(self, value):
        if value < 0:
            raise ValueError
        else:
            self._epoch = value

    @property
    def kind(self):
        '''str: Message kind'''
        return self._kind

    @kind.setter
    def kind(self, value):
        if value not in ALLOWED_KINDS:
            raise ValueError(f'Invalid message kind "{value}"')
        else:
            self._kind = value

    @property
    def data(self):
        '''torch.Tensor, tuple(torch.Tensor) or None:  This field is optional for
        :class:`MetaMessage`, but mandatory for :class:`TrainingMessage`'''
        return self._data

    @abc.abstractproperty
    def key(self):
        '''object: A key used for grouping messages into :class:`MessageBundle` s'''
        pass

    def __str__(self):
        return (f'<{self.__class__.__name__}: seq={self.seq}, '
                f'step={self.step}, epoch={self.epoch}, kind={self.kind}>')

    def __repr__(self):
        return str(self)


class MetaMessage(Message):
    '''A message with meta information not tied to any specific module. Can still carry tensor data,
    if necessary.'''
    def __init__(self, tag, seq, step, epoch, kind, data=None):
        super().__init__(tag, seq, step, epoch, kind)
        self._data = data

    @Message.kind.setter
    def kind(self, value):
        if value not in META_KINDS:
            raise ValueError(f'Invalid message kind "{value}"')
        else:
            self._kind = value

    @property
    def data(self):
        '''torch.Tensor, tuple or None: Optional data. Can be used e.g. for input to the network,
        labels or network output'''
        return self._data

    @property
    def key(self):
        return 'META'


class TrainingMessage(Message):
    '''A message tied to a specific module, with tensor data attached.'''

    def __init__(self, tag, seq, step, epoch, kind, module, data):
        super().__init__(tag, seq, step, epoch, kind)
        self._module  = module
        if data is None:
            raise ValueError('Data cannot be `None` for `TrainingMessage`')
        self._data = data

    @property
    def module(self):
        '''torch.nn.Module: Module emitting this data'''
        return self._module

    @Message.kind.setter
    def kind(self, value):
        if value not in DATA_KINDS:
            raise ValueError(f'Invalid message kind "{value}"')
        else:
            self._kind = value

    @property
    def key(self):
        return self.module.name


class MessageBundle(object):

    '''Data object for holding a set of artifacts for a module (or meta information) at one point
    during training. This data type can be used to buffer different kinds and check whether all
    expected kinds have been received for a module or meta information.

    .. note::

        This docstring doesn't make sense yet.

    '''
    def __init__(self, identifier, kinds):
        '''
        Parameters
        ----------
        identifier  :   str
                        String to uniquely identify the bundle. Will usually be either the module
                        name or `'META'` for a collection of meta messages.
        kinds   :   list
                    Message kinds contained in this bundle
        '''
        if isinstance(kinds, str):
            # this has bitten me before. base `Subscription`s don't use multiple kinds
            kinds = [kinds]
        self._identifier     = identifier
        self._expected_kinds = kinds
        self._received       = {kind: False for kind in kinds}
        self._data           = {kind: None for kind in kinds}
        self._seq            = None
        self._step           = None
        self._epoch          = None

    @property
    def identifier(self):
        ''' str: A string denoting the common aspect of the collected messages (besides the step).
        This can be the module name or a string such as ``META`` or other  denoting these are
        messages which do not belong to a module.'''
        return self._identifier

    @property
    def expected_kinds(self):
        '''list(str): The expected kinds of messages per iteration '''
        return self._expected_kinds

    @property
    def data(self):
        '''dict(str, torch.Tensor): The tensors received for each kind'''
        return self._data

    @property
    def seq(self):
        '''int: Global sequence number of this class'''
        return self._seq

    @property
    def step(self):
        '''int: Sequence number (training step) of the received messages (should match across all
        msgs in one iteration)'''
        return self._step

    def epoch(self):
        '''int: Epoch of the received messages (should match across all msgs in one iteration)'''
        return self._epoch

    def complete(self):
        '''Check if all expected messages have been received. This means the bundle can be released
        to subscribers.

        Returns
        -------
        bool
        '''
        return all(self._received.values())

    def _check_message(self, message):
        '''Check consistency of sequence number, step and epoch or set if not set yet. Check
        consistency of identifier and check for duplication.

        Parameters
        ----------
        message :   ikkuna.export.messages.Message

        Raises
        ------
        ValueError
            If ``message.(seq|step|epoch|identifier)`` does not match the current
            ``(seq|step|epoch|identifier)`` or in case a message of ``message.kind`` has already
            been received
        '''
        ############
        #  seqnum  #
        ############
        if not self._seq:
            self._seq = message.seq

        if message.seq != self._seq:
            raise ValueError(f'Attempting to add message with seq {message.seq} to bundle with '
                             'initial seq {self._seq}')

        ##########
        #  step  #
        ##########
        if not self._step:
            self._step = message.step

        if message.step != self._step:
            raise ValueError(f'Attempting to add message with step {message.step} to bundle with '
                             'initial step {self._step}')
        #############
        #  content  #
        #############

        if not self._epoch:
            self._epoch = message.epoch

        if message.epoch != self._epoch:
            raise ValueError(f'Attempting to add message from epoch {message.epoch} to bundle with '
                             'initial epoch {self._epoch}')

        ################
        #  Identifier  #
        ################
        if self._identifier != message.key:
            raise ValueError(f'Unexpected identifier "{message.key}" (expected "{self._identifier}")')

        #################
        #  Duplication  #
        #################
        if self._received[message.kind]:
            raise ValueError(f'Got duplicate value for kind "{message.kind}".')

    def add_message(self, message):
        '''Add a new message to this object. Will fail if the new messsage does not have the same
        sequence number and epoch.

        Parameters
        ----------
        message :   ikkuna.export.messages.Message

        Raises
        ------
        ValueError
            see :meth:`~MessageBundle._check_message()`
        '''
        self._check_message(message)

        self._received[message.kind] = True

        if message.data is not None:
            self._data[message.kind] = message.data

    def __getattr__(self, name):
        '''Override to mimick a property for each kind of message in this data (e.g.
        ``message_bundle.activations`` instead of ``message_bundle.data['activations']``)
        '''
        if name in self._expected_kinds:
            return self._data[name]
        else:
            return self.__getattribute__(name)

    def __str__(self):
        mod   = self._identifier
        step  = self._step
        kinds = list(self._data.keys())
        return f'<MessageBundle: identifier={mod}, kinds={kinds}, step={step}>'

    def __repr__(self):
        return str(self)
